#include "wmbus.h"
#include <drivers/uart.h>
#define LOG_MODULE_NAME wmbus
LOG_MODULE_REGISTER(LOG_MODULE_NAME, LOG_LEVEL_DBG);


#define UART_WAIT_FOR_BUF_DELAY K_MSEC(50)



static K_MEM_SLAB_DEFINE(uart_slab, BUF_SIZE, 100, 4); // can store 100 wmbus frames.
static K_MEM_SLAB_DEFINE(payload_slab, BUF_SIZE, 100, 4); // can store 100 wmbus frames.




static K_FIFO_DEFINE(m_wmbus_frame_tx); // stor the whole frame in FIFO manner
static K_FIFO_DEFINE(m_wmbus_frame_rx); // store the whole RX Frame in FIFO Manner
static K_FIFO_DEFINE(fifo_uart_tx_data);
static K_FIFO_DEFINE(fifo_uart_rx_data);


static struct k_delayed_work uart_work_rx_data;


/* NOTE: After K_FIFO_GET free the memory of payload as well. */


/* @brief UART Callback
 */

static void uart_callback(const struct device *dev,
			  struct uart_event *evt,
			  void *user_data)
{
	//struct device *uart = user_data;
	struct uart_data_t *buffer; // it would be a pointer to the event buffer
	static size_t aborted_buffer_len;
	static uint8_t *aborted_buffer = NULL;
	int err;

	switch (evt->type) {
	case UART_TX_DONE:
		LOG_INF("Tx sent %d bytes", evt->data.tx.len);
		
		// This executes when a previous aborted buffer is succesfully transferred
		if (aborted_buffer) {
			// Get a pointer to a container structure from an element.
			buffer = CONTAINER_OF(aborted_buffer, struct uart_data_t, data);
			aborted_buffer = NULL;
			aborted_buffer_len = 0;
		}
		else {
			//Get a pointer to a container structure from an element.
			buffer = CONTAINER_OF(evt->data.tx.buf, struct uart_data_t, data);
		}
		// free the allocated memory for tx
		k_mem_slab_free(&uart_slab, (void **)&buffer);

		// get next data to send
		buffer = k_fifo_get(&fifo_uart_tx_data, K_NO_WAIT);
		// if there is no remaining data to send then return
		if (!buffer) {
			return;
		}
		// send the data if there is any in the fifo_uart_tx_data
		if (uart_tx(dev, buffer->data, buffer->len, SYS_FOREVER_MS)) {
				LOG_WRN("Failed to send data over UART");
		}
			
		break;

	case UART_TX_ABORTED:
		LOG_ERR("Tx aborted");
		if (!aborted_buffer) {
			aborted_buffer = (uint8_t *)evt->data.tx.buf;
		}
		aborted_buffer_len += evt->data.tx.len;
		buffer = CONTAINER_OF(aborted_buffer, struct uart_data_t, data);

		uart_tx(dev, &buffer->data[aborted_buffer_len], buffer->len - aborted_buffer_len, SYS_FOREVER_MS);
		break;

	case UART_RX_RDY:
		LOG_INF("Received data %d bytes", evt->data.rx.len);
		buffer = CONTAINER_OF(evt->data.rx.buf, struct uart_data_t, data);
		buffer->len += evt->data.rx.len;
		k_fifo_put(&fifo_uart_rx_data, buffer);

		break;

	case UART_RX_BUF_REQUEST:
	{
		struct uart_data_t *buf;

		err = k_mem_slab_alloc(&uart_slab, (void **)&buf, K_NO_WAIT);
		__ASSERT(err == 0, "Failed to allocate slab");

		err = uart_rx_buf_rsp(dev, buf->data, sizeof(buf->data));
		__ASSERT(err == 0, "Failed to provide new buffer");
		break;
	}

	case UART_RX_BUF_RELEASED:
	 	// do not release data here unless the K_fifo_get is called and then it will be released.
		//k_mem_slab_free(&uart_slab, (void **)&evt->data.rx_buf.buf);
		k_delayed_work_submit(&uart_work_rx_data, UART_WAIT_FOR_BUF_DELAY);
		break;

	case UART_RX_DISABLED:
		break;

	case UART_RX_STOPPED:
		break;
	}
}

static uint8_t calculate_checksum(uint8_t* pdata, size_t len)
{
	uint8_t cs;
	for(uint8_t n = 0; n <len; n++){
		cs ^= pdata[n];
	}

	return cs;
}

void metis_frame_print(struct metis_frame_t *frame)
{
	uint8_t n=0;
	LOG_INF("SFD:     %02X\n", frame->SFD);
	LOG_INF("CMD:     %02X\n", frame->CMD);
	LOG_INF("LEN:     %02X\n", frame->len);
	for(n=0; n<frame->len; n++){
    	LOG_INF("P[%03d]:  %02X\n", n, frame->payload[n]);
	}
	LOG_INF("CS:      %02X\n", frame->cs);
}

static void uart_work_rx_data_hdl(struct k_work *item)
{
	int err;
	struct uart_data_t* rx_buffer = k_fifo_get(&m_wmbus_frame_rx, K_NO_WAIT);
	struct metis_frame_t frame;
	frame.SFD = rx_buffer->data[0];
	frame.CMD = rx_buffer->data[1];
	frame.len = rx_buffer->data[2];
	frame.payload=NULL;
	frame.cs = calculate_checksum(rx_buffer->data, (frame.len + 3));
	if(frame.SFD != METIS_SFD) {
		LOG_ERR("Wrong start of frame");
		goto out; 
	};

	if(rx_buffer->len-3 <= frame.len) {
		LOG_ERR("incomplete payload received");
		goto out;
	};

	err = k_mem_slab_alloc(&payload_slab, (void **)&frame.payload, K_NO_WAIT);
	__ASSERT(err == 0, "Failed to allocate slab");
	memcpy(&rx_buffer->data[3], frame.payload, frame.len);

	k_fifo_put(&m_wmbus_frame_rx, &frame);

	out:
	k_mem_slab_free(&uart_slab, (void **)&rx_buffer); // at the end release memory because this is already been free by uart driver
}


int metis_frame_serialize(const struct device *uart, wmbus_cmd_t cmd, uint8_t * payload, size_t len) 
{
	struct uart_data_t* uart_tx_data;
	k_mem_slab_alloc(&uart_slab, (void **)&uart_tx_data, K_NO_WAIT);

	if(!uart_tx_data){
		LOG_ERR("Out of memory");
		return -ENOMEM;
	}
	uart_tx_data->data[0]=METIS_SFD;
	uart_tx_data->data[1]=(uint8_t)cmd;
	uart_tx_data->data[2]=len;
	if(len != 0) {
		memcpy(&(uart_tx_data->data[3]), payload, len);
	}
	uart_tx_data->data[3+len]=calculate_checksum(uart_tx_data->data, (3+len));
	if (uart_tx(uart, uart_tx_data->data, (3+len), SYS_FOREVER_MS)) {
		LOG_WRN("Failed to send data over UART");
	}
	return 0;
}

void wmbus_rx_thread(void)
{
	uint32_t rx_count = 0;
	for (;;) {
		/* Wait indefinitely for new data from wm-bus module */
		struct metis_frame_t* rx_frame = k_fifo_get(&m_wmbus_frame_rx,
						     K_FOREVER);

		LOG_INF("data received: %d", rx_count++);
		metis_frame_print(rx_frame);
		//LOG_HEXDUMP_INF(buf->data, buf->len, "");

		// free the payload
		k_mem_slab_free(&payload_slab, (void **)&rx_frame->payload);
	}
}


void memtis_init(const struct device *uart)
{

	int err;
	struct uart_data_t *buf;

	err = k_mem_slab_alloc(&uart_slab, (void **)&buf, K_NO_WAIT);
	LOG_INF("SUCCESS alloc slab %d\n", err);
	__ASSERT(err == 0, "Failed to alloc slab");

	err = uart_callback_set(uart, uart_callback, NULL);
	LOG_INF("SUCCESS to set callback %d\n", err);
	//__ASSERT(err == 0, "Failed to set callback");

	err = uart_rx_enable(uart, buf->data, sizeof(buf->data), 1000);
	LOG_INF("Failed to enable RXk %d\n", err);
	__ASSERT(err == 0, "Failed to enable RX");

	LOG_INF("uart rx enable sucessfully\n");
	k_delayed_work_init(&uart_work_rx_data, uart_work_rx_data_hdl);
	
}

void memtis_enable_deepsleep(const struct device *uart)
{
	uint8_t payload = 0x08;
	metis_frame_serialize(uart, CMD_SETUARTSPEED_REQ, &payload, 1);
}

void memtis_wakeup(const struct device *uart)
{
	uint8_t payload = 0x03;
	metis_frame_serialize(uart, CMD_SETUARTSPEED_REQ, &payload, 1);
}

void memtis_get_serial_number(const struct device *uart)
{
	metis_frame_serialize(uart, CMD_SETUARTSPEED_REQ, NULL, 0);
}
